#include <string.h>
#include <gtk/gtk.h>

#include "cfg.h"

#include "guiutils.h"
#include "stacklist.h"
#include "keymap_list.h"
#include "cfg_win.h"

#include "edv_types.h"
#include "libendeavour2-base/edv_utils.h"
#include "edv_utils_gtk.h"
#include "archiver.h"
#include "customize_win.h"
#include "endeavour2.h"

#include "edv_cfg_list.h"
#include "config.h"


/* Callbacks */
static void EDVCustomizeArchiverKeymapsDefaultsCB(
	keymap_list_struct *kmlist, keymap_key_struct *km,
	gpointer data
);

/* Create Pages */
void EDVCustomizeArchiverCreateKeymapsPage(
	EDVCore *core,
	GtkWidget *cfgwin,
	GtkWidget *parent
);
void EDVCustomizeArchiverCreateContentsListPage(
	EDVCore *core,
	GtkWidget *cfgwin,
	GtkWidget *parent
);
void EDVCustomizeArchiverCreateToolBarPage(
	EDVCore *core,
	GtkWidget *cfgwin,
	GtkWidget *parent
);
void EDVCustomizeArchiverCreateAppearancePage(
	EDVCore *core,
	GtkWidget *cfgwin,
	GtkWidget *parent
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (gint)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *	Keymaps List defaults callback.
 */
static void EDVCustomizeArchiverKeymapsDefaultsCB(
	keymap_list_struct *kmlist, keymap_key_struct *km,
	gpointer data
)
{
	GList *glist;
	CfgAccelkey *ak;
	CfgAccelkeyList *ak_list;
	EDVArchiverOpID id;
	EDVCore *core = EDV_CORE(data);
	if((kmlist == NULL) || (km == NULL) || (core == NULL))
		return;

	id = (EDVArchiverOpID)km->data;
	ak_list = edv_archiver_accelkeys_new();
	if(ak_list == NULL)
		return;

	for(glist = ak_list->list;
	    glist != NULL;
	    glist = g_list_next(glist)
	)
	{
		ak = CFG_ACCELKEY(glist->data);
		if(ak == NULL)
			continue;

		if(ak->op_id == id)
		{
			km->keyval = ak->key;
			km->state = ak->modifiers;
			break;
		}
	}    
	 
	CFGAccelkeyListDelete(ak_list);
}

/*
 *	Creates the Archiver Keymaps page.
 */
void EDVCustomizeArchiverCreateKeymapsPage(
	EDVCore *core,
	GtkWidget *cfgwin,
	GtkWidget *parent
)
{
	const gint      border_major = 5,
			border_minor = 2;
	GList *op_ids_list;
	GtkWidget	*w,
			*parent2;
	CfgList *cfg_list = core->cfg_list;
	CfgAccelkeyList *ak_list;
	CfgWinLink *cfg_link;
	keymap_list_struct *kmlist;

	gtk_container_border_width(GTK_CONTAINER(parent), border_major);

	/* Accelerator Keys GtkFrame */
	w = gtk_frame_new(
#if defined(PROG_LANGUAGE_SPANISH)
"Acelerador Adapta"
#elif defined(PROG_LANGUAGE_FRENCH)
"L'Accelerateur Saisit"
#elif defined(PROG_LANGUAGE_GERMAN)
"Beschleuniger Gibt Ein"
#elif defined(PROG_LANGUAGE_ITALIAN)
"Chiavi Di Acceleratore"
#elif defined(PROG_LANGUAGE_DUTCH)
"De Versneller Stemt"
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"Teclas De Acelerador"
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Akselerator Fester"
#else
"Accelerator Keys"
#endif
);
	gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 0);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_hbox_new(FALSE, border_minor);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_widget_show(w);
	parent2 = w;

	/* Create the Keymap List */
	kmlist = KeymapListNew(
		parent2,
		GTK_ORIENTATION_HORIZONTAL
	);
	KeymapListSetEditable(kmlist, TRUE);
	KeymapListSetAutoCheckConflicts(kmlist, TRUE);
	KeymapListSetKeyDefaultRequestCB(
		kmlist,
		EDVCustomizeArchiverKeymapsDefaultsCB,
		core
	);
	KeymapListMap(kmlist);
	cfg_link = CfgWinAddLink(
		cfgwin,
		parent,
		CFG_WIN_LINK_WIDGET_KEYMAP_LIST,
		kmlist,
		EDV_CFG_PARM_ARCHIVER_ACCELERATOR_KEYS
	);

	/* Add the Accelkeys to the Keymap List */
	KeymapListClear(kmlist);
	op_ids_list = edv_archiver_op_ids_list_new(NULL);
	ak_list = EDV_GET_ACCELKEY_LIST(
		EDV_CFG_PARM_ARCHIVER_ACCELERATOR_KEYS
	);
	if(ak_list != NULL)
	{
		GList	*glist,
			*glist2,
			*ak_glist = ak_list->list;
		CfgAccelkey *ak;
		EDVArchiverOpID id;
		EDVArchiverOp *op;

		/* Iterate through the Operation IDs */
		for(glist = op_ids_list;
		    glist != NULL;
		    glist = g_list_next(glist)
		)
		{
			op = EDV_ARCHIVER_OP(glist->data);
			if(op == NULL)
				continue;

			if(op->flags & EDV_OPID_NO_OP)
				continue;

			/* Get the EDVArchiverOpID */
			id = op->id;

			/* Find an Accelkey for this Operation ID */
			for(glist2 = ak_glist;
			    glist2 != NULL;
			    glist2 = g_list_next(glist2)
			)
			{
				ak = CFG_ACCELKEY(glist2->data);
				if(ak == NULL)
					continue;

				if(ak->op_id == (gint)id)
				{
					/* Add this Accelkey to the Keymap List */
					gchar *name = EDVCustomizeSimplifyOPIDName(
						op->menu_name
					);
					KeymapListAppend(
						kmlist,
						op->button_icon_data,
						name,
						KeymapKeyNew(
							ak->key,
							ak->modifiers,
							(gpointer)id
						)
					);
					g_free(name);
					break;
				}
			}
		}
	}
	edv_archiver_op_ids_list_delete(op_ids_list);
}


/*
 *	Create Archiver Contents List page.
 */
void EDVCustomizeArchiverCreateContentsListPage(
	EDVCore *core,
	GtkWidget *cfgwin,
	GtkWidget *parent
)
{
	const gint	border_major = 5,
					border_minor = 2;
	GtkWidget *w, *parent2, *parent3;
	CfgWinLink *cfg_link;
	stack_list_struct *slist;

	gtk_container_border_width(GTK_CONTAINER(parent), border_major);

	/* Cell Display Options GtkFrame */
	w = gtk_frame_new(
#if defined(PROG_LANGUAGE_SPANISH)
"Las Opciones Del Despliegue De La Clula"
#elif defined(PROG_LANGUAGE_FRENCH)
"Les Options D'Exposition De Cellule"
#elif defined(PROG_LANGUAGE_GERMAN)
"Zelle Ausstellung Optionen"
#elif defined(PROG_LANGUAGE_ITALIAN)
"Le Opzioni Di Mostra Di Cellula"
#elif defined(PROG_LANGUAGE_DUTCH)
"Cel Tentoonstelling Opties"
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"As Opes De Exposio De Clula"
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Celleut Stillings Alternativer"
#else
"Cell Display Options"
#endif
	);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_hbox_new(FALSE, border_minor);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_widget_show(w);
	parent2 = w;

	/* Hide Directory Size GtkCheckButton */
	w = gtk_check_button_new_with_label(
#if defined(PROG_LANGUAGE_SPANISH)
"Esconda Tamao Gua"
#elif defined(PROG_LANGUAGE_FRENCH)
"Cacher Taille D'Annuaire"
#elif defined(PROG_LANGUAGE_GERMAN)
"Verstecken Sie Verzeichnis Gre"
#elif defined(PROG_LANGUAGE_ITALIAN)
"La Misura Di Elenco Di Nascondiglio"
#elif defined(PROG_LANGUAGE_DUTCH)
"Verberg Gids Maat"
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"Esconda Tamanho De Guia"
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Gjemm Katalog Strrelse"
#else
"Hide Directory Size"
#endif
	);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	cfg_link = CfgWinAddLink(
		cfgwin,
		parent,
		CFG_WIN_LINK_WIDGET_TOGGLE_BUTTON,
		w,
		EDV_CFG_PARM_ARCHIVER_CONTENTS_HIDE_DIR_SIZE
	);

	/* Hide Link Size GtkCheckButton */
	w = gtk_check_button_new_with_label(
#if defined(PROG_LANGUAGE_SPANISH)
"Esconda tamao del enlace"
#elif defined(PROG_LANGUAGE_FRENCH)
"Cacher Taille Lien"
#elif defined(PROG_LANGUAGE_GERMAN)
"Verstecken Sie Verbindung Gre"
#elif defined(PROG_LANGUAGE_ITALIAN)
"La Misura Di Maglia Di Nascondiglio"
#elif defined(PROG_LANGUAGE_DUTCH)
"Verberg Schakel Maat"
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"O Tamanho De Elo De Couro"
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Gjemm LeddStrrelse"
#else
"Hide Link Size"
#endif
	);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	cfg_link = CfgWinAddLink(
		cfgwin,
		parent,
		CFG_WIN_LINK_WIDGET_TOGGLE_BUTTON,
		w,
		EDV_CFG_PARM_ARCHIVER_CONTENTS_HIDE_LINK_SIZE
	);

	/* Hide Link Permission GtkCheckButton */
	w = gtk_check_button_new_with_label(
#if defined(PROG_LANGUAGE_SPANISH)
"Esconda permisos del enlace"
#elif defined(PROG_LANGUAGE_FRENCH)
"Cacher Permissions Lien"
#elif defined(PROG_LANGUAGE_GERMAN)
"Verstecken Sie Verbindung Erlaubnis"
#elif defined(PROG_LANGUAGE_ITALIAN)
"I Permessi Di Maglia Di Nascondiglio"
#elif defined(PROG_LANGUAGE_DUTCH)
"Verberg Schakel Toestemmingen"
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"As Permisses De Elo De Couro"
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Gjemm Ledd Tillatelse"
#else
"Hide Link Permissions"
#endif
	);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	cfg_link = CfgWinAddLink(
		cfgwin,
		parent,
		CFG_WIN_LINK_WIDGET_TOGGLE_BUTTON,
		w,
		EDV_CFG_PARM_ARCHIVER_CONTENTS_HIDE_LINK_PERMISSIONS
	);


	/* Arrangement GtkFrame */
	w = gtk_frame_new(
#if defined(PROG_LANGUAGE_SPANISH)
"El Arreglo"
#elif defined(PROG_LANGUAGE_FRENCH)
"Arrangement"
#elif defined(PROG_LANGUAGE_GERMAN)
"Anordnung"
#elif defined(PROG_LANGUAGE_ITALIAN)
"La Disposizione"
#elif defined(PROG_LANGUAGE_DUTCH)
"Schikking"
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"O Arranjo"
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Arrangement"
#else
"Arrangement"
#endif
	);
	gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 0);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_vbox_new(FALSE, border_minor);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	/* Automatically Resize Columns GtkCheckButton */
	w = gtk_check_button_new_with_label(
"Automatically Resize Columns"
	);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	GUISetWidgetTip(
		w,
"Check this to automatically resize the columns each time the list\
 is updated"
	);
	gtk_widget_show(w);
	cfg_link = CfgWinAddLink(
		cfgwin,
		parent,
		CFG_WIN_LINK_WIDGET_TOGGLE_BUTTON,
		w,
		EDV_CFG_PARM_ARCHIVER_CONTENTS_LIST_AUTO_RESIZE_COLUMNS
	);

	/* Contents List Stack List */
	slist = StackListNew(
		parent2,
#if defined(PROG_LANGUAGE_SPANISH)
"Disponible", "La Corriente"
#elif defined(PROG_LANGUAGE_FRENCH)
"Disponible", "Courant"
#elif defined(PROG_LANGUAGE_GERMAN)
"Verfgbar", "Strmung"
#elif defined(PROG_LANGUAGE_ITALIAN)
"Disponibile", "La Corrente"
#elif defined(PROG_LANGUAGE_DUTCH)
"Verkrijgbaar", "Stroom"
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"Disponvel", "A Corrente"
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Tilgjengelig", "Strm"
#else
"Available", "Current"
#endif
	);
	StackListShowDescription(slist, TRUE, TRUE);
	StackListSetMaintainSourceOrder(slist, TRUE);
	StackListMap(slist);
	cfg_link = CfgWinAddLink(
		cfgwin,
		parent,
		CFG_WIN_LINK_WIDGET_STACK_LIST,
		slist,
		EDV_CFG_PARM_ARCHIVER_CONTENTS_COLUMN
	);

	/* Add the headings list */
	StackListAppend(
		slist,
"Index",
"Archive object index number",
		NULL,			/* No icon data */
		NULL,			/* No user data */
		EDV_ARCHIVER_COLUMN_TYPE_INDEX,
		FALSE, FALSE		/* No multiple, no stay on target */
	);
	StackListAppend(
		slist,
#if defined(PROG_LANGUAGE_SPANISH)
"El Nombre",
"El nombre del objeto",
#elif defined(PROG_LANGUAGE_FRENCH)
"Nom",
"Le nom de l'objet",
#elif defined(PROG_LANGUAGE_GERMAN)
"Name",
"Name des objekts",
#elif defined(PROG_LANGUAGE_ITALIAN)
"Il Nome",
"Il nome dell'oggetto",
#elif defined(PROG_LANGUAGE_DUTCH)
"Naam",
"Naam van het voorwerp",
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"O Nome",
"O nome do objeto",
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Navn",
"Navn av objektet",
#else
"Name",
"Name of the object",
#endif
		NULL,			/* No icon data */
		NULL,			/* No user data */
		EDV_ARCHIVER_COLUMN_TYPE_NAME,
		FALSE, TRUE			/* No multiple, stay on target */
	);
	StackListAppend(
		slist,
#if defined(PROG_LANGUAGE_SPANISH)
"El Tamao",
"El tamao del objeto",
#elif defined(PROG_LANGUAGE_FRENCH)
"Taille",
"La taille de l'objet",
#elif defined(PROG_LANGUAGE_GERMAN)
"Gre.",
"Gre des objekts",
#elif defined(PROG_LANGUAGE_ITALIAN)
"La Misura",
"La misura dell'oggetto",
#elif defined(PROG_LANGUAGE_DUTCH)
"Maat",
"Maat van het voorwerp",
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"O Tamanho",
"O tamanho do objeto",
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Strrelse",
"Strrelse av objektet",
#else
"Size",
"Original size of the object",
#endif
		NULL,			/* No icon data */
		NULL,			/* No user data */
		EDV_ARCHIVER_COLUMN_TYPE_SIZE,
		FALSE, FALSE		/* No multiple, no stay on target */
	);
	StackListAppend(
		slist,
#if defined(PROG_LANGUAGE_SPANISH)
"El Tamao Comprimido",
"El tamao comprimido del objeto",
#elif defined(PROG_LANGUAGE_FRENCH)
"Taille Serre",
"La taille serre de l'objet",
#elif defined(PROG_LANGUAGE_GERMAN)
"Zusammengedrckte Gre",
"Zusammengedrckte gre des objekts",
#elif defined(PROG_LANGUAGE_ITALIAN)
"La Misura Compressa",
"La misura compressa dell'oggetto",
#elif defined(PROG_LANGUAGE_DUTCH)
"Samengedrukte Maat",
"Samengedrukte maat van het voorwerp",
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"Tamanho Comprimido",
"Tamanho comprimido do objeto",
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Komprimert Strrelse",
"Komprimert strrelse av objektet",
#else
"Storage Size",
"Size the object takes up in the archive",
#endif
		NULL,			/* No icon data */
		NULL,			/* No user data */
		EDV_ARCHIVER_COLUMN_TYPE_STORAGE_SIZE,
		FALSE, FALSE		/* No multiple, no stay on target */
	);
	StackListAppend(
		slist,
#if defined(PROG_LANGUAGE_SPANISH)
"El Tipo",
"De tipo MIME se asoci con el objeto",
#elif defined(PROG_LANGUAGE_FRENCH)
"Type",
"Le Type de MIME a associ avec l'objet",
#elif defined(PROG_LANGUAGE_GERMAN)
"Typ",
"MIME Typ hat verbunden mit dem objekt",
#elif defined(PROG_LANGUAGE_ITALIAN)
"Il Tipo",
"Il tipo di MIME  frequentato l'oggetto",
#elif defined(PROG_LANGUAGE_DUTCH)
"Type",
"MIME type associeerde met het voorwerp",
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"O Tipo",
"O tipo de MIME associou-se com o objeto",
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Type",
"MIME maskinskriver tilknyttet med objektet",
#else
"Type",
"MIME Type associated with the object",
#endif
		NULL,			/* No icon data */
		NULL,			/* No user data */
		EDV_ARCHIVER_COLUMN_TYPE_TYPE,
		FALSE, FALSE		/* No multiple, no stay on target */
	);
	StackListAppend(
		slist,
#if defined(PROG_LANGUAGE_SPANISH)
"Los Permisos",
"Los permisos para el objeto",
#elif defined(PROG_LANGUAGE_FRENCH)
"Permissions",
"Les permissions pour l'objet",
#elif defined(PROG_LANGUAGE_GERMAN)
"Erlaubnis",
"Erlaubnis fr das objekt",
#elif defined(PROG_LANGUAGE_ITALIAN)
"I Permessi",
"I permessi per l'oggetto",
#elif defined(PROG_LANGUAGE_DUTCH)
"Toestemmingen",
"Toestemmingen voor het voorwerp",
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"As Permisses",
"As permisses para o objeto",
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Tillatelse",
"Tillatelse for objektet",
#else
"Permissions",
"Access rights to the object",
#endif
		NULL,			/* No icon data */
		NULL,			/* No user data */
		EDV_ARCHIVER_COLUMN_TYPE_PERMISSIONS,
		FALSE, FALSE		/* No multiple, no stay on target */
	);
	StackListAppend(
		slist,
#if defined(PROG_LANGUAGE_SPANISH)
"El Dueo",
"El dueo del objeto",
#elif defined(PROG_LANGUAGE_FRENCH)
"Propritaire",
"Le propritaire de l'objet",
#elif defined(PROG_LANGUAGE_GERMAN)
"Eigentmer",
"Eigentmer des objekts",
#elif defined(PROG_LANGUAGE_ITALIAN)
"Il Proprietario",
"Il proprietario dell'oggetto",
#elif defined(PROG_LANGUAGE_DUTCH)
"Eigenaar",
"Eigenaar van het voorwerp",
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"O Proprietrio",
"O proprietrio do objeto",
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Eier",
"Eier av objektet",
#else
"Owner",
"User that the object belongs to",
#endif
		NULL,			/* No icon data */
		NULL,			/* No user data */
		EDV_ARCHIVER_COLUMN_TYPE_OWNER,
		FALSE, FALSE		/* No multiple, no stay on target */
	);
	StackListAppend(
		slist,
#if defined(PROG_LANGUAGE_SPANISH)
"El Grupo",
"La propiedad agrupada del objeto",
#elif defined(PROG_LANGUAGE_FRENCH)
"Groupe",
"La possession groupe de l'objet",
#elif defined(PROG_LANGUAGE_GERMAN)
"Gruppe",
"Klassifiziertes eigentumsrecht des objekts",
#elif defined(PROG_LANGUAGE_ITALIAN)
"Il Gruppo",
"La propriet raggruppata dell'oggetto",
#elif defined(PROG_LANGUAGE_DUTCH)
"Groep",
"Gegroepeerd eigendomsrecht van het voorwerp",
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"O Grupo",
"Posse agrupada do objeto",
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Gruppe",
"Gruppert eiendomsrett av objektet",
#else
"Group",
"Group that the object belongs to",
#endif
		NULL,			/* No icon data */
		NULL,			/* No user data */
		EDV_ARCHIVER_COLUMN_TYPE_GROUP,
		FALSE, FALSE		/* No multiple, no stay on target */
	);
	StackListAppend(
		slist,
#if defined(PROG_LANGUAGE_SPANISH)
"El Acceso De La Fecha",
"Feche que el objeto era ltimo conseguir acceso a",
#elif defined(PROG_LANGUAGE_FRENCH)
"Accs De Date",
"Dater que l'objet tait dernier a accd ",
#elif defined(PROG_LANGUAGE_GERMAN)
"Datieren Sie Zugriff",
"Datum, da das objekt zuletzt auf zugegriffen wurde",
#elif defined(PROG_LANGUAGE_ITALIAN)
"L'Accesso Di Data",
"Data che l'oggetto era ultimo accesso a",
#elif defined(PROG_LANGUAGE_DUTCH)
"Dateer Toegang",
"Datum dat het voorwerp was vorig toegang tot had",
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"O Acesso De Data",
"Date que o objeto era ltimo acedido",
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Dater Tilgang",
"Dato at objektet var siste kommet til",
#else
"Date Access",
"Date and time that the object was last read from or written to",
#endif
		NULL,			/* No icon data */
		NULL,			/* No user data */
		EDV_ARCHIVER_COLUMN_TYPE_DATE_ACCESS,
		FALSE, FALSE		/* No multiple, no stay on target */
	);
	StackListAppend(
		slist,
#if defined(PROG_LANGUAGE_SPANISH)
"Feche Modificado",
"La fecha que el objeto era ltimo modificado",
#elif defined(PROG_LANGUAGE_FRENCH)
"Dater Modifi",
"La date que l'objet tait dernier",
#elif defined(PROG_LANGUAGE_GERMAN)
"Datum Hat Modifiziert",
"Datum, da das objekt zuletzt modifiziert wurde",
#elif defined(PROG_LANGUAGE_ITALIAN)
"Datare Modificato",
"Data che l'oggetto era ultimo modificato",
#elif defined(PROG_LANGUAGE_DUTCH)
"Datum Wijzigde",
"Datum dat het voorwerp vorig gewijzigde was",
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"Date Modificado",
"Data que o objeto era ltimos modificados",
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Dato Modifisert",
"Dato at objektet var siste modifisert",
#else
"Date Modified",
"Date and time that the object was last written to",
#endif
		NULL,			/* No icon data */
		NULL,			/* No user data */
		EDV_ARCHIVER_COLUMN_TYPE_DATE_MODIFIED,
		FALSE, FALSE		/* No multiple, no stay on target */
	);
	StackListAppend(
		slist,
#if defined(PROG_LANGUAGE_SPANISH)
"Feche Cambiado",
"Feche que el objeto era ltimo cambiado",
#elif defined(PROG_LANGUAGE_FRENCH)
"Dater Chang",
"Dater que l'objet tait dernier",
#elif defined(PROG_LANGUAGE_GERMAN)
"Datum Hat Gendert",
"Datum, da das objekt zuletzt gendert wurde",
#elif defined(PROG_LANGUAGE_ITALIAN)
"Datare Cambiato",
"Data che l'oggetto era ultimo cambiato",
#elif defined(PROG_LANGUAGE_DUTCH)
"Datum Veranderde",
"Datum dat het voorwerp vorig veranderde was",
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"Date Mudado",
"Date que o objeto era ltimo mudado",
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Dato Forandret Seg",
"Dato at objektet var siste forandret",
#else
"Date Changed",
"Date and time that the object was last written to or its properties were modified",
#endif
		NULL,			/* No icon data */
		NULL,			/* No user data */
		EDV_ARCHIVER_COLUMN_TYPE_DATE_CHANGED,
		FALSE, FALSE		/* No multiple, no stay on target */
	);
	StackListAppend(
		slist,
#if defined(PROG_LANGUAGE_SPANISH)
"La Ubicacin",
"La ubicacin para el objeto cuando extrado",
#elif defined(PROG_LANGUAGE_FRENCH)
"Emplacement",
"L'emplacement pour l'objet quand extrait",
#elif defined(PROG_LANGUAGE_GERMAN)
"Ort",
"Ort fr das objekt wenn extrahiert",
#elif defined(PROG_LANGUAGE_ITALIAN)
"La Posizione",
"La posizione per l'oggetto quando estratto",
#elif defined(PROG_LANGUAGE_DUTCH)
"Plaats",
"Plaats voor het voorwerp wanneer onttrekkenene",
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"A Localidade",
"A localidade para o objeto quando extrado",
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Plassering",
"Plassering for objektet nr trukket ut",
#else
"Location",
"Location for the object when extracted",
#endif
		NULL,			/* No icon data */
		NULL,			/* No user data */
		EDV_ARCHIVER_COLUMN_TYPE_LOCATION,
		FALSE, FALSE		/* No multiple, no stay on target */
	);
	StackListAppend(
		slist,
#if defined(PROG_LANGUAGE_SPANISH)
"Ligado A",
"Ligue el destino",
#elif defined(PROG_LANGUAGE_FRENCH)
"Reli A",
"Relier la destination",
#elif defined(PROG_LANGUAGE_GERMAN)
"Verknpft Zu",
"Verknpfen sie reiseziel",
#elif defined(PROG_LANGUAGE_ITALIAN)
"Concatenato A",
"Concatenare la destinazione",
#elif defined(PROG_LANGUAGE_DUTCH)
"Verbindenene Te",
"Verbind bestemming",
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"Ligado A",
"Ligue destino",
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Forbundet Til",
"Forbind destinasjon",
#else
"Linked To",
"Link target value",
#endif
		NULL,			/* No icon data */
		NULL,			/* No user data */
		EDV_ARCHIVER_COLUMN_TYPE_LINKED_TO,
		FALSE, FALSE		/* No multiple, no stay on target */
	);
	StackListAppend(
		slist,
#if defined(PROG_LANGUAGE_SPANISH)
"De Tipo Dispositivo",
"El dispositivo el par mayor y secundario del nmero",
#elif defined(PROG_LANGUAGE_FRENCH)
"Type D'Appareil",
"Le commandant d'appareil et la paire de numro mineure",
#elif defined(PROG_LANGUAGE_GERMAN)
"Gertetyp",
"Gertemajor und minderjhriger zhlen paar",
#elif defined(PROG_LANGUAGE_ITALIAN)
"Il Tipo Di Congegno",
"Congegno maggiore e minore numero",
#elif defined(PROG_LANGUAGE_DUTCH)
"Apparaat Type",
"Apparaat majoor en minderjarige nummeren paar",
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"O Tipo De Artifcio",
"O artifcio par menor importante de nmero",
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Innretnings Type",
"Innretnings major og mindr erigantall par",
#else
"Device Type",
"Device major and minor number pair",
#endif
		NULL,			/* No icon data */
		NULL,			/* No user data */
		EDV_ARCHIVER_COLUMN_TYPE_DEVICE_TYPE,
		FALSE, FALSE		/* No multiple, no stay on target */
	);
	StackListAppend(
		slist,
"Encryption",
"Name of the encryption algorithm or blank if not encrypted",
		NULL,			/* No icon data */
		NULL,			/* No user data */
		EDV_ARCHIVER_COLUMN_TYPE_ENCRYPTION,
		FALSE, FALSE		/* No multiple, no stay on target */
	);
	StackListAppend(
		slist,
#if defined(PROG_LANGUAGE_SPANISH)
"La Compresin",
"La proporcin de la compresin de los datos de objeto",
#elif defined(PROG_LANGUAGE_FRENCH)
"Compression",
"La proportion de compression des donnes de l'objet",
#elif defined(PROG_LANGUAGE_GERMAN)
"Kompression",
"Kompression verhltnis die daten des des objekts",
#elif defined(PROG_LANGUAGE_ITALIAN)
"La Compressione",
"l rapporto di compressione dei dati dell'oggetto",
#elif defined(PROG_LANGUAGE_DUTCH)
"Samendrukking",
"Samendrukking verhouding van het voorwerp's de gegevens",
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"A Compresso",
"A relao de compresso dos dados do objeto",
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Kompresjon",
"Kompresjonsforhold av objektets data",
#else
"Compression",
"Compression ratio of the object's data",
#endif
		NULL,			/* No icon data */
		NULL,			/* No user data */
		EDV_ARCHIVER_COLUMN_TYPE_COMPRESSION,
		FALSE, FALSE		/* No multiple, no stay on target */
	);
	StackListAppend(
		slist,
"Method",
"Name of the storage method",
		NULL,			/* No icon data */
		NULL,			/* No user data */
		EDV_ARCHIVER_COLUMN_TYPE_METHOD,
		FALSE, FALSE		/* No multiple, no stay on target */
	);
	StackListAppend(
		slist,
#if defined(PROG_LANGUAGE_SPANISH)
"CRC",
"Checksum de CRC de los datos de objeto",
#elif defined(PROG_LANGUAGE_FRENCH)
"CRC",
"CRC checksum des donnes de l'objet",
#elif defined(PROG_LANGUAGE_GERMAN)
"CRC",
"CRC checksum die Daten des des Objekts",
#elif defined(PROG_LANGUAGE_ITALIAN)
"CRC",
"Il totale di verifica di CRC dei dati dell'oggetto",
#elif defined(PROG_LANGUAGE_DUTCH)
"CRC",
"CRC checksum van het voorwerp's de gegevens",
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"CRC",
"O checksum de CRC dos dados do objeto",
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"CRC",
"CRC checksum av objektets data",
#else
"CRC",
"Cyclic Redundancy Check value",
#endif
		NULL,			/* No icon data */
		NULL,			/* No user data */
		EDV_ARCHIVER_COLUMN_TYPE_CRC,
		FALSE, FALSE		/* No multiple, no stay on target */
	);

	StackListUpdateMenus(slist);

}

/*
 *      Creates the Archiver Tool Bar page.
 */
void EDVCustomizeArchiverCreateToolBarPage(
	EDVCore *core,
	GtkWidget *cfgwin,
	GtkWidget *parent
)
{
	const gint	border_major = 5,
			border_minor = 2;
	gchar *name;
	GList		*glist,
			*op_ids_list;
	GSList *gslist;
	GtkWidget	*w,
			*parent2, *parent3;
	CfgWinLink *cfg_link;
	stack_list_struct *slist;
	EDVOpIDFlags flags;
	EDVArchiverOp *op;

	gtk_container_border_width(GTK_CONTAINER(parent), border_major);

	/* Display Buttons As and Buttons Relief GtkHBox */
	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	/* Display Buttons As GtkFrame */
	w = gtk_frame_new(
#if defined(PROG_LANGUAGE_SPANISH)
"El Despliegue Abrocha Como"
#elif defined(PROG_LANGUAGE_FRENCH)
"L'exposition Boutonne Comme"
#elif defined(PROG_LANGUAGE_GERMAN)
"Ausstellung Knpft Als Zu"
#elif defined(PROG_LANGUAGE_ITALIAN)
"La Mostra Abbottona Come"
#elif defined(PROG_LANGUAGE_DUTCH)
"Tentoonstelling Knoopt Als Dicht"
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"Exiba Botes Como"
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Utstilling Knapper Som"
#else
"Display Buttons As"
#endif
	);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_widget_show(w);
	parent3 = w;

	w = gtk_hbox_new(FALSE, border_minor);
	gtk_container_add(GTK_CONTAINER(parent3), w);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_widget_show(w);
	parent3 = w;

	/* Text GtkRadioButton */
	gslist = NULL;
	w = gtk_radio_button_new_with_label(
		gslist,
#if defined(PROG_LANGUAGE_SPANISH)
"El Texto"
#elif defined(PROG_LANGUAGE_FRENCH)
"Texte"
#elif defined(PROG_LANGUAGE_GERMAN)
"Text"
#elif defined(PROG_LANGUAGE_ITALIAN)
"Il Testo"
#elif defined(PROG_LANGUAGE_DUTCH)
"Tekst"
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"O Texto"
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Tekst"
#else
"Text"
#endif
	);
	gslist = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	cfg_link = CfgWinAddLink(
		cfgwin,
		parent,
		CFG_WIN_LINK_WIDGET_RADIO_BUTTON,
		w,
		EDV_CFG_PARM_ARCHIVER_TOOL_BAR_BUTTON_DECALS
	);
	CfgWinLinkSetData(
		cfg_link,
		(gpointer)TOOL_BAR_BUTTON_DECALS_TEXT
	);

	/* Pictures GtkRadioButton */
	w = gtk_radio_button_new_with_label(
		gslist,
#if defined(PROG_LANGUAGE_SPANISH)
"Los Retratos"
#elif defined(PROG_LANGUAGE_FRENCH)
"Images"
#elif defined(PROG_LANGUAGE_GERMAN)
"Bilder"
#elif defined(PROG_LANGUAGE_ITALIAN)
"Le Immagini"
#elif defined(PROG_LANGUAGE_DUTCH)
"Afbeelding"
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"Os Quadros"
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Bilder"
#else
"Pictures"
#endif
	);
	gslist = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	cfg_link = CfgWinAddLink(
		cfgwin,
		parent,
		CFG_WIN_LINK_WIDGET_RADIO_BUTTON,
		w,
		EDV_CFG_PARM_ARCHIVER_TOOL_BAR_BUTTON_DECALS
	);
	CfgWinLinkSetData(
		cfg_link,
		(gpointer)TOOL_BAR_BUTTON_DECALS_PICTURES
	);

	/* Pictures & Text GtkRadioButton */
	w = gtk_radio_button_new_with_label(
		gslist,
#if defined(PROG_LANGUAGE_SPANISH)
"Los Retratos & Texto"
#elif defined(PROG_LANGUAGE_FRENCH)
"Les Images & Le Texte"
#elif defined(PROG_LANGUAGE_GERMAN)
"Bilder & Text"
#elif defined(PROG_LANGUAGE_ITALIAN)
"Le Immagini & Il Testo"
#elif defined(PROG_LANGUAGE_DUTCH)
"Afbeelding & Tekst"
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"Os Quadros & Texto"
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Bilder & Tekst"
#else
"Pictures & Text"
#endif
	);
/*	gslist = gtk_radio_button_group(GTK_RADIO_BUTTON(w)); */
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	cfg_link = CfgWinAddLink(
		cfgwin,
		parent,
		CFG_WIN_LINK_WIDGET_RADIO_BUTTON,
		w,
		EDV_CFG_PARM_ARCHIVER_TOOL_BAR_BUTTON_DECALS
	);
	CfgWinLinkSetData(
		cfg_link,
		(gpointer)TOOL_BAR_BUTTON_DECALS_PICTURES_AND_TEXT
	);


	/* Buttons Relief GtkFrame */
	w = gtk_frame_new(
#if defined(PROG_LANGUAGE_SPANISH)
"El Alivio De Botones"
#elif defined(PROG_LANGUAGE_FRENCH)
"Soulagement De Boutons"
#elif defined(PROG_LANGUAGE_GERMAN)
"Knpfe Erleichterung"
#elif defined(PROG_LANGUAGE_ITALIAN)
"Il Sollievo Di Bottoni"
#elif defined(PROG_LANGUAGE_DUTCH)
"Knopen Steun"
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"O Alvio De Botes"
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Knapp Lettelse"
#else
"Buttons Relief"
#endif
	);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_widget_show(w);
	parent3 = w;

	w = gtk_hbox_new(FALSE, border_minor);
	gtk_container_add(GTK_CONTAINER(parent3), w);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_widget_show(w);
	parent3 = w;

	/* Flush GtkRadioButton */
	gslist = NULL;
	w = gtk_radio_button_new_with_label(
		gslist,
#if defined(PROG_LANGUAGE_SPANISH)
"El Rubor"
#elif defined(PROG_LANGUAGE_FRENCH)
"Eclat"
#elif defined(PROG_LANGUAGE_GERMAN)
"Splen"
#elif defined(PROG_LANGUAGE_ITALIAN)
"Il Flusso"
#elif defined(PROG_LANGUAGE_DUTCH)
"Vloed"
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"Nivele"
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Spyl"
#else
"Flush"
#endif
	);
	gslist = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	cfg_link = CfgWinAddLink(
		cfgwin,
		parent,
		CFG_WIN_LINK_WIDGET_RADIO_BUTTON,
		w,
		EDV_CFG_PARM_ARCHIVER_TOOL_BAR_RELIEF
	);
	CfgWinLinkSetData(
		cfg_link,
		(gpointer)GTK_RELIEF_NONE
	);

	/* Half GtkRadioButton */
	w = gtk_radio_button_new_with_label(
		gslist,
#if defined(PROG_LANGUAGE_SPANISH)
"La Mitad"
#elif defined(PROG_LANGUAGE_FRENCH)
"Moiti"
#elif defined(PROG_LANGUAGE_GERMAN)
"Hlfte"
#elif defined(PROG_LANGUAGE_ITALIAN)
"La Met"
#elif defined(PROG_LANGUAGE_DUTCH)
"Helft"
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"A Metade"
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Halvdel"
#else
"Half"
#endif
	);
	gslist = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	cfg_link = CfgWinAddLink(
		cfgwin,
		parent,
		CFG_WIN_LINK_WIDGET_RADIO_BUTTON,
		w,
		EDV_CFG_PARM_ARCHIVER_TOOL_BAR_RELIEF
	);
	CfgWinLinkSetData(
		cfg_link,
		(gpointer)GTK_RELIEF_HALF
	);

	/* Raised GtkRadioButton */
	w = gtk_radio_button_new_with_label(
		gslist,
#if defined(PROG_LANGUAGE_SPANISH)
"Levantado"
#elif defined(PROG_LANGUAGE_FRENCH)
"Hauss"
#elif defined(PROG_LANGUAGE_GERMAN)
"Erhoben"
#elif defined(PROG_LANGUAGE_ITALIAN)
"Aumentato"
#elif defined(PROG_LANGUAGE_DUTCH)
"Opgetilde"
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"Levantado"
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Hevet"
#else
"Raised"
#endif
	);
/*	gslist = gtk_radio_button_group(GTK_RADIO_BUTTON(w)); */
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	cfg_link = CfgWinAddLink(
		cfgwin,
		parent,
		CFG_WIN_LINK_WIDGET_RADIO_BUTTON,
		w,
		EDV_CFG_PARM_ARCHIVER_TOOL_BAR_RELIEF
	);
	CfgWinLinkSetData(
		cfg_link,
		(gpointer)GTK_RELIEF_NORMAL
	);


	/* Arrangement GtkFrame */
	w = gtk_frame_new(
#if defined(PROG_LANGUAGE_SPANISH)
"El Arreglo"
#elif defined(PROG_LANGUAGE_FRENCH)
"Arrangement"
#elif defined(PROG_LANGUAGE_GERMAN)
"Anordnung"
#elif defined(PROG_LANGUAGE_ITALIAN)
"La Disposizione"
#elif defined(PROG_LANGUAGE_DUTCH)
"Schikking"
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"O Arranjo"
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Arrangement"
#else
"Arrangement"
#endif
	);
	gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 0);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_vbox_new(FALSE, border_minor);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_widget_show(w);
	parent2 = w;

	/* Tool Bar Stack List */
	slist = StackListNew(
		parent2,
#if defined(PROG_LANGUAGE_SPANISH)
"Disponible", "La Corriente"
#elif defined(PROG_LANGUAGE_FRENCH)
"Disponible", "Courant"
#elif defined(PROG_LANGUAGE_GERMAN)
"Verfgbar", "Strmung"
#elif defined(PROG_LANGUAGE_ITALIAN)
"Disponibile", "La Corrente"
#elif defined(PROG_LANGUAGE_DUTCH)
"Verkrijgbaar", "Stroom"
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"Disponvel", "A Corrente"
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Tilgjengelig", "Strm"
#else
"Available", "Current"
#endif
	);
	StackListShowDescription(slist, TRUE, TRUE);
	StackListSetMaintainSourceOrder(slist, TRUE);
	StackListMap(slist);
	cfg_link = CfgWinAddLink(
		cfgwin,
		parent,
		CFG_WIN_LINK_WIDGET_STACK_LIST,
		slist,
		EDV_CFG_PARM_ARCHIVER_TOOL_BAR
	);

	/* Add the tool bar items from the Archiver operation ids */
	op_ids_list = edv_archiver_op_ids_list_new(NULL);
	for(glist = op_ids_list;
	    glist != NULL;
	    glist = g_list_next(glist)
	)
	{
		op = EDV_ARCHIVER_OP(glist->data);
		if(op == NULL)
			continue;

		flags = op->flags;
		if(flags & EDV_OPID_EX_TOOL_BAR)
			continue;

		name = EDVCustomizeSimplifyOPIDName(op->menu_name);
		StackListAppend(
			slist,
			name,
			op->tooltip,
			op->button_icon_data,
			NULL,				/* No user data */
			(gint)op->id,			/* ID */
			(flags & EDV_OPID_ALLOW_MULTIPLE) ? TRUE : FALSE,
			FALSE				/* Not stay on target */
		);
		g_free(name);
	}
	edv_archiver_op_ids_list_delete(op_ids_list);

	StackListUpdateMenus(slist);
}

/*
 *	Creates the Archiver Appearance page.
 */
void EDVCustomizeArchiverCreateAppearancePage(
	EDVCore *core,
	GtkWidget *cfgwin,
	GtkWidget *parent
)
{
	const gint	border_major = 5,
					border_minor = 2;
	GtkWidget	*w,
					*parent2, *parent3, *parent4, *parent5;
	CfgWinLink *cfg_link;

	gtk_container_border_width(GTK_CONTAINER(parent), border_major);

	/* GtkHBox to separate two columns */
	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;


	/* First column GtkVBox */
	w = gtk_vbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent3 = w;

	/* Show In Title Bar GtkFrame */
	w = gtk_frame_new(
#if defined(PROG_LANGUAGE_SPANISH)
"Muestre En La Barra Del Ttulo"
#elif defined(PROG_LANGUAGE_FRENCH)
"Montrer Dans La Barre De Titre"
#elif defined(PROG_LANGUAGE_GERMAN)
"Zeigen Sie In Titel Stab"
#elif defined(PROG_LANGUAGE_ITALIAN)
"Mostrare Nella Sbarra Di Titolo"
#elif defined(PROG_LANGUAGE_DUTCH)
"Toon In Titel Staaf"
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"Mostre Em Barra De Ttulo"
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Vis In Title Bar"
#else
"Show In Title Bar"
#endif
	);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_widget_show(w);
	parent4 = w;

	w = gtk_vbox_new(FALSE, border_minor);
	gtk_container_add(GTK_CONTAINER(parent4), w);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_widget_show(w);
	parent4 = w;

	/* Show Program Name */
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent5 = w;
	/* GtkCheckButton */
	w = gtk_check_button_new_with_label("Program Name");
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	cfg_link = CfgWinAddLink(
		cfgwin,
		parent,
		CFG_WIN_LINK_WIDGET_TOGGLE_BUTTON,
		w,
		EDV_CFG_PARM_ARCHIVER_TITLE_WITH_PROGRAM
	);

	/* Show User Name */
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent5 = w;
	/* GtkCheckButton */
	w = gtk_check_button_new_with_label(
#if defined(PROG_LANGUAGE_SPANISH)
"El Nombre Del Usuario"
#elif defined(PROG_LANGUAGE_FRENCH)
"Nom D'Utilisateur"
#elif defined(PROG_LANGUAGE_GERMAN)
"Verbraucher Name"
#elif defined(PROG_LANGUAGE_ITALIAN)
"Il Nome Di Operatore"
#elif defined(PROG_LANGUAGE_DUTCH)
"Gebruiker Naam"
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"O Nome De Operador"
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Bruker Name"
#else
"User Name"
#endif
	);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	cfg_link = CfgWinAddLink(
		cfgwin,
		parent,
		CFG_WIN_LINK_WIDGET_TOGGLE_BUTTON,
		w,
		EDV_CFG_PARM_ARCHIVER_TITLE_WITH_USER
	);

	/* Show Location */
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent5 = w;
	/* GtkCheckButton */
	w = gtk_check_button_new_with_label(
#if defined(PROG_LANGUAGE_SPANISH)
"La Ubicacin"
#elif defined(PROG_LANGUAGE_FRENCH)
"Emplacement"
#elif defined(PROG_LANGUAGE_GERMAN)
"Ort"
#elif defined(PROG_LANGUAGE_ITALIAN)
"La Posizione"
#elif defined(PROG_LANGUAGE_DUTCH)
"Plaats"
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"A Localidade"
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Plassering"
#else
"Location"
#endif
	);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	cfg_link = CfgWinAddLink(
		cfgwin,
		parent,
		CFG_WIN_LINK_WIDGET_TOGGLE_BUTTON,
		w,
		EDV_CFG_PARM_ARCHIVER_TITLE_WITH_LOCATION
	);

	/* Reverse Order */
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent5 = w;
	/* GtkCheckButton */
	w = gtk_check_button_new_with_label("Reverse Order");
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	cfg_link = CfgWinAddLink(
		cfgwin,
		parent,
		CFG_WIN_LINK_WIDGET_TOGGLE_BUTTON,
		w,
		EDV_CFG_PARM_ARCHIVER_TITLE_REVERSE
	);


	/* Display GtkFrame */
	w = gtk_frame_new(
#if defined(PROG_LANGUAGE_SPANISH)
"El Despliegue"
#elif defined(PROG_LANGUAGE_FRENCH)
"Exposition"
#elif defined(PROG_LANGUAGE_GERMAN)
"Ausstellung"
#elif defined(PROG_LANGUAGE_ITALIAN)
"La Mostra"
#elif defined(PROG_LANGUAGE_DUTCH)
"Tentoonstelling"
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"A Exposio"
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Utstilling"
#else
"Display"
#endif
	);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_widget_show(w);
	parent4 = w;

	w = gtk_vbox_new(FALSE, border_minor);
	gtk_container_add(GTK_CONTAINER(parent4), w);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_widget_show(w);
	parent4 = w;

	/* Show Tool Bar */
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent5 = w;
	/* GtkCheckButton */
	w = gtk_check_button_new_with_label(
#if defined(PROG_LANGUAGE_SPANISH)
"Barra De Instrumento"
#elif defined(PROG_LANGUAGE_FRENCH)
"Barre D'Outil"
#elif defined(PROG_LANGUAGE_GERMAN)
"Werkzeug Stab"
#elif defined(PROG_LANGUAGE_ITALIAN)
"La Sbarra Di attrezzo"
#elif defined(PROG_LANGUAGE_DUTCH)
"Werktuig Staaf"
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"A Barra De Ferramenta"
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Redskap Bar"
#else
"Tool Bar"
#endif
	);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	cfg_link = CfgWinAddLink(
		cfgwin,
		parent,
		CFG_WIN_LINK_WIDGET_TOGGLE_BUTTON,
		w,
		EDV_CFG_PARM_ARCHIVER_SHOW_TOOL_BAR
	);

	/* Show Location Bar */
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent5 = w;
	/* GtkCheckButton */
	w = gtk_check_button_new_with_label(
#if defined(PROG_LANGUAGE_SPANISH)
"Barra De Ubicacin"
#elif defined(PROG_LANGUAGE_FRENCH)
"Barre D'Emplacement"
#elif defined(PROG_LANGUAGE_GERMAN)
"Ort Stab"
#elif defined(PROG_LANGUAGE_ITALIAN)
"La Sbarra Di Posizione"
#elif defined(PROG_LANGUAGE_DUTCH)
"Plaats Staaf"
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"A Barra De Localidade"
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Plassering Bar"
#else
"Location Bar"
#endif
	);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	cfg_link = CfgWinAddLink(
		cfgwin,
		parent,
		CFG_WIN_LINK_WIDGET_TOGGLE_BUTTON,
		w,
		EDV_CFG_PARM_ARCHIVER_SHOW_LOCATION_BAR
	);

	/* Show Find Bar */
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent5 = w;
	/* GtkCheckButton */
	w = gtk_check_button_new_with_label(
#if defined(PROG_LANGUAGE_SPANISH)
"Barra La Encuentre"
#elif defined(PROG_LANGUAGE_FRENCH)
"Barre La Trouver"
#elif defined(PROG_LANGUAGE_GERMAN)
"Stab Sie Finden"
#elif defined(PROG_LANGUAGE_ITALIAN)
"La Sbarra Trova"
#elif defined(PROG_LANGUAGE_DUTCH)
"Sluit Vondst Uit"
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"A Barra Acha"
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Sperr Find"
#else
"Find Bar"
#endif
	);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	cfg_link = CfgWinAddLink(
		cfgwin,
		parent,
		CFG_WIN_LINK_WIDGET_TOGGLE_BUTTON,
		w,
		EDV_CFG_PARM_ARCHIVER_SHOW_FIND_BAR
	);

	/* Show Status Bar */
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent5 = w;
	/* GtkCheckButton */
	w = gtk_check_button_new_with_label(
#if defined(PROG_LANGUAGE_SPANISH)
"Barra La Posicin"
#elif defined(PROG_LANGUAGE_FRENCH)
"Barre La Statut"
#elif defined(PROG_LANGUAGE_GERMAN)
"Stab Sie Status"
#elif defined(PROG_LANGUAGE_ITALIAN)
"La Sbarra Lo Stato"
#elif defined(PROG_LANGUAGE_DUTCH)
"Sluit Status"
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"A Barra Estado"
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Sperr Status"
#else
"Status Bar"
#endif
	);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	cfg_link = CfgWinAddLink(
		cfgwin,
		parent,
		CFG_WIN_LINK_WIDGET_TOGGLE_BUTTON,
		w,
		EDV_CFG_PARM_ARCHIVER_SHOW_STATUS_BAR
	);


	/* Second column GtkVBox */
	w = gtk_vbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent3 = w;

	/* Archive Display GtkFrame */
	w = gtk_frame_new(
#if defined(PROG_LANGUAGE_SPANISH)
"El Despliegue Del Archivo"
#elif defined(PROG_LANGUAGE_FRENCH)
"Archiver L'Exposition"
#elif defined(PROG_LANGUAGE_GERMAN)
"Archiv Ausstellung"
#elif defined(PROG_LANGUAGE_ITALIAN)
"La Mostra Di Archivio"
#elif defined(PROG_LANGUAGE_DUTCH)
"Archief Tentoonstelling"
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"A Exposio De Arquivo"
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Arkiv Display"
#else
"Archive Display"
#endif
	);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_widget_show(w);
	parent4 = w;

	w = gtk_vbox_new(FALSE, border_minor);
	gtk_container_add(GTK_CONTAINER(parent4), w);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_widget_show(w);
	parent4 = w;

	/* Show Archive Info On Open */
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent5 = w;
	/* GtkCheckButton */
	w = gtk_check_button_new_with_label(
"Show Archive Comments On Open"
	);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	GUISetWidgetTip(
		w,
"Check this to display the archive's comments when opening an archive\
 that contains comments"
	);
	gtk_widget_show(w);
	cfg_link = CfgWinAddLink(
		cfgwin,
		parent,
		CFG_WIN_LINK_WIDGET_TOGGLE_BUTTON,
		w,
		EDV_CFG_PARM_ARCHIVER_SHOW_ARCHIVE_COMMENTS_ON_OPEN
	);
}
